package at.ac.uibk.gitsearch.repository.search;

import java.io.IOException;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.StringTokenizer;
import java.util.Timer;
import java.util.TimerTask;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Consumer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.annotation.PostConstruct;
import javax.ws.rs.NotFoundException;

import org.apache.commons.lang3.StringUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultsDTO;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO.Person;
import org.codeability.sharing.plugins.api.search.util.ExerciseId;
import org.elasticsearch.ElasticsearchException;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.ExistsQueryBuilder;
import org.elasticsearch.index.query.MultiMatchQueryBuilder;
import org.elasticsearch.index.query.QueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.index.query.TermQueryBuilder;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.elasticsearch.search.fetch.subphase.highlight.HighlightBuilder;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.context.event.ContextStoppedEvent;
import org.springframework.context.event.EventListener;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.userdetails.User;
import org.springframework.stereotype.Repository;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;

import at.ac.uibk.gitsearch.domain.Authority;
import at.ac.uibk.gitsearch.service.dto.AutoCompleteEntry;

@Repository
@SuppressWarnings("PMD.GodClass")
public class MetaDataRepository {

	private final RestHighLevelClient elasticsearchClient;

	private static final Logger LOGGER = LogManager.getLogger(MetaDataRepository.class);

	private static final int FRAGMENT_SIZE = 5000;

	private static final long AUTOCOMPLETION_RECHECK_TIMEOUT = 5 * 60 * 1000L; // every 5 Minutes
	private Timer autocompletionReloadTimer = new Timer("AutocompletionReloadTimer");

	public static final String EXERCISE_BY_ID_CACHE = "at.ac.uibk.gitsearch.repository.search.MetaDataRepository.ExerciseByIdCache";

	@Value("${application.search.highlight-pre}")
	private String preTags;

	@Value("${application.search.highlight-post}")
	private String postTags;

	/**
	 * contains for each Field, a list of Strings mapped to a completion. Together
	 * with hit counts E.g. "metadata.creator" -> Kerstin -> {"Kerstin Limbeck" ->
	 * 3, "Kerstin Mair" -> 2)}
	 */
	private Map<String, Map<String, Map<String, Integer>>> cachedCompletions = null;

	public MetaDataRepository(RestHighLevelClient elasticsearchClient) {
		this.elasticsearchClient = elasticsearchClient;
	}

	/**
	 * returns all keyword autocompletes for keyWord
	 *
	 * @param keyWordPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getKeywordsAutoComplete(String keyWordPrefix, int max) throws IOException {
		return getFieldAutoCompletion(keyWordPrefix, SearchRepositoryConstants.METADATA_KEYWORDS, max);
	}

	/**
	 * returns all keyword autocompletes for keyWord
	 *
	 * @param formatPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getFormatsAutoComplete(String formatPrefix, int max) throws IOException {
		return getFieldAutoCompletion(formatPrefix, SearchRepositoryConstants.METADATA_FORMATS, max);
	}

	/**
	 * returns all programmingLanguage autocompletes for keyWord
	 *
	 * @param programmingLanguagePrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getProgrammingLanguageAutoComplete(String programmingLanguagePrefix, int max)
			throws IOException {
		return getFieldAutoCompletion(programmingLanguagePrefix,
				SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES, max);
	}

	/**
	 * returns all creator autocompletes
	 *
	 * @param creatorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getCreatorAutoComplete(String creatorPrefix, int max) throws IOException {
		return getFieldAutoCompletion(creatorPrefix, SearchRepositoryConstants.METADATA_CREATOR, max);
	}

	/**
	 * returns all contributor autocompletes
	 *
	 * @param contributorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getContributorAutoComplete(String contributorPrefix, int max) throws IOException {
		return getFieldAutoCompletion(contributorPrefix, SearchRepositoryConstants.METADATA_CONTRIBUTOR, max);
	}

	/**
	 * returns all contributor and creator autocompletes
	 *
	 * @param contributorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getContributorCreatorAutoComplete(String contributorPrefix, int max)
			throws IOException {
		final List<AutoCompleteEntry> contributors = getContributorAutoComplete(contributorPrefix, max);
		final List<AutoCompleteEntry> creatorAutoComplete = getCreatorAutoComplete(contributorPrefix, max);
		// merging
		for (AutoCompleteEntry creator : creatorAutoComplete) {
			contributors.stream().filter(entry -> entry.getTarget().equals(creator.getTarget())).findAny()
					.ifPresentOrElse(entry -> {
						contributors.remove(entry);
						contributors.add(new AutoCompleteEntry(creator.getTarget(),
								creator.getHitCount() + entry.getHitCount()));
					}, () -> contributors.add(creator));
		}
		return contributors;
	}

	/**
	 * returns a list of hits together with the number of occurences
	 *
	 * @param prefix        the prefix
	 * @param metaDataField the meta data field
	 * @param maxCount      the maximum of hits, returned
	 * @return
	 * @throws IOException
	 * @throws JsonProcessingException
	 * @throws JsonMappingException
	 */
	private List<AutoCompleteEntry> getFieldAutoCompletion(String prefix, final String metaDataField, int maxCount)
			throws IOException {
		fillAutoCompletion();

		String lcKeyWordPrefix = prefix.toLowerCase(Locale.getDefault());
		// höllisch kompliziert und höllisch unverständlich mit Streams :-)
		// idea
		// 1. Filtern nach Prefix
		// 2. Zusammenzählen der möglichen Hits (Besser Max statt Sum?)
		// 3. Auf MAX_AUTO_COMPLETION_RESULTS begrenzen und nur die Trefferstrings
		// ausgeben.
		final Stream<Entry<String, Map<String, Integer>>> filteredEntries = cachedCompletions.get(metaDataField)
				.entrySet().stream() // Map<String, Map<String, Integer>> -> Stream<Entry<String, Map<String,
				// Integer>>
				.filter(s -> s.getKey().startsWith(lcKeyWordPrefix));
		final Map<String, Integer> combinedHits = filteredEntries // s = String -> bool -> Stream<Entry<String,
				// Map<String, Integer>>
				.flatMap(e -> e.getValue().entrySet().stream()) // Stream< Map<String, Integer>> with duplicates
				.collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue, Integer::max));
		return combinedHits.entrySet().stream().sorted((e1, e2) -> e2.getValue() - e1.getValue()).limit(maxCount)
				.map(s -> new AutoCompleteEntry(s.getKey(), s.getValue())).collect(Collectors.toList()); // Map.Entry<String,
		// Map<String,
		// Integer>>

	}

	@PostConstruct
	@SuppressWarnings("PMD")
	private void setUpAutocompletion() {
		autocompletionReloadTimer.schedule(new TimerTask() {
			@Override
			public void run() {
				try {
					updateAutocompletionCache();
				} catch (IOException | RuntimeException e) {
					LOGGER.warn("Cannot refresh Autocompletion Cache", e);
				}
			}
		}, 0L, AUTOCOMPLETION_RECHECK_TIMEOUT);
	}

	/**
	 * stops the cache reload timer, during shutdown or between tests.
	 *
	 * @param event ContextStoppedEvent
	 */
	@EventListener
	protected void stopReloadTimer(ContextStoppedEvent event) {
		autocompletionReloadTimer.cancel();
	}

	private synchronized void fillAutoCompletion() throws IOException {
		synchronized (this) {
			if (cachedCompletions == null) {
				updateAutocompletionCache();
			}
		}
	}

	private void updateAutocompletionCache() throws IOException {
		Map<String, Map<String, Map<String, Integer>>> reloadedCachedCompletions = new ConcurrentHashMap<>();
		reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_KEYWORDS, new TreeMap<>());
		reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_CREATOR, new TreeMap<>());
		reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_CONTRIBUTOR, new TreeMap<>());
		reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES, new TreeMap<>());
		reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_FORMATS, new TreeMap<>());

		SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_METADATA);

		final int pageSize = 20;
		int currentPage = 0;
		boolean tryNextPage = true;

		while (tryNextPage) {
			SearchSourceBuilder sourceBuilder = new SearchSourceBuilder().size(pageSize).from(currentPage++ * pageSize);

			searchRequest.source(sourceBuilder);

			SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);

			ObjectMapper objectMapper = getSearchResultObjectMapper();

			tryNextPage = searchResponse.getHits().getHits().length > 0;

			for (SearchHit searchHit : searchResponse.getHits().getHits()) {
				final String sourceAsJSON = searchHit.getSourceAsString();
				SearchResultDTO entry = objectMapper.readValue(sourceAsJSON, SearchResultDTO.class);
				if (entry.getMetadata().getKeyword() != null) {
					for (String keyWord : entry.getMetadata().getKeyword())
						addTo(reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_KEYWORDS),
								split(keyWord), keyWord);
				}
				if (entry.getMetadata().getFormat() != null) {
					for (String format : entry.getMetadata().getFormat()) {
						addTo(reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_FORMATS), split(format),
								format);}
				}
				if (entry.getMetadata().getContributor() != null) {
					for (Person contributor : entry.getMetadata().getContributor()) {
						addTo(reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_CONTRIBUTOR),
								split(contributor.getName()), contributor.getName());
					}
				}
				if (entry.getMetadata().getCreator() != null) {
					for (Person creator : entry.getMetadata().getCreator()) {
						addTo(reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_CREATOR),
								split(creator.getName()), creator.getName());
					}
				}
				if (entry.getMetadata().getProgrammingLanguage() != null) {
					for (String language : entry.getMetadata().getProgrammingLanguage()) {
						addTo(reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES),
								Collections.singletonList(language), language);
					}
				}
			}
		}
		cachedCompletions = reloadedCachedCompletions;
	}

	/**
	 * returns a preconfigured object mapper
	 * 
	 * @return
	 */
	public static ObjectMapper getSearchResultObjectMapper() {
		ObjectMapper objectMapper = new ObjectMapper();
		objectMapper.registerModule(new JavaTimeModule());
		objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
		return objectMapper;
	}

	private static void addTo(Map<String, Map<String, Integer>> completionMap, List<String> tokenList, String target) {
		tokenList.forEach(token -> {
			String lowercaseToken = token.toLowerCase(Locale.getDefault());
			Map<String, Integer> existingTargets = completionMap.get(lowercaseToken);
			if (existingTargets == null) { // none found
				existingTargets = new HashMap<>();
				existingTargets.put(target, 1);
				completionMap.put(lowercaseToken, existingTargets);
			} else {
				final Integer count = existingTargets.get(target);
				if (count == null) {
					existingTargets.put(target, 1);
				} else {
					existingTargets.put(target, count + 1);
				}
			}
		});
	}

	private static List<String> split(String s) {
		StringTokenizer st = new StringTokenizer(s, " \t\n\r\f");
		List<String> result = new ArrayList<>();
		while (st.hasMoreTokens()) {
			result.add(st.nextToken());
		}
		return result;
	}

	public SearchResultsDTO searchByEmail(String email, int pageSize, Optional<User> user) throws IOException {
		SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_METADATA);

		BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery();

		BoolQueryBuilder authorBuilder = QueryBuilders.boolQuery();
		if (email != null) {
			authorBuilder.should(QueryBuilders
					.multiMatchQuery(email, SearchRepositoryConstants.METADATA_CREATOR_EMAIL,
							SearchRepositoryConstants.METADATA_CONTRIBUTOR_EMAIL,
							SearchRepositoryConstants.METADATA_PUBLISHER_EMAIL)
					.type(MultiMatchQueryBuilder.Type.PHRASE_PREFIX));
		}
		if (authorBuilder.hasClauses()) {
			queryBuilder.must(authorBuilder);
		}

		// Authorization restrictions
		final BoolQueryBuilder publicQuery = createPublicQuery();

		if (user.isEmpty()) {
			queryBuilder.must(publicQuery);
		} else {
			final Collection<GrantedAuthority> authorities = user.get().getAuthorities();
			final BoolQueryBuilder authQuery = QueryBuilders.boolQuery().boost(0.0f);
			final Stream<QueryBuilder> prefixAuthQueries = authorities.stream()
					.filter(auth -> !auth.getAuthority().startsWith("ROLE"))
					.filter(auth -> !auth.getAuthority().startsWith("SCOPE")).map(auth -> QueryBuilders
							.prefixQuery(SearchRepositoryConstants.PROJECT_NAMESPACE, auth.getAuthority()));
			prefixAuthQueries.forEach(authQuery::should);
			if (authQuery.hasClauses()) {
				queryBuilder.must(QueryBuilders.boolQuery().should(authQuery).should(publicQuery));
			} else {
				queryBuilder.must(publicQuery);
			}
		}

		SearchSourceBuilder sourceBuilder = new SearchSourceBuilder();
		sourceBuilder.query(queryBuilder).size(pageSize).from(0);

		searchRequest.source(sourceBuilder);

		SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);

		return parseSearchResponse(searchResponse);
	}

	/**
	 * returns the hits from the meta data search index for the input query
	 * parameters.
	 *
	 * @param searchInputDTO the query parameters.
	 * @param pageSize       the size of the page (i.e. the maximum of returned
	 *                       hits).
	 * @param user           the user that queries.
	 * @return the search hits.
	 * @throws IOException when search index is not available.
	 */
	public SearchResultsDTO pageDetails(org.codeability.sharing.plugins.api.search.SearchInputDTO searchInputDTO,
			Optional<User> user) throws IOException {
		fillAutoCompletion();

		BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery();

		BoolQueryBuilder fullTextBuilder = QueryBuilders.boolQuery();
		forEachTerm(searchInputDTO.getFulltextQuery(),
				term -> fullTextBuilder.should(QueryBuilders
						.multiMatchQuery(term, SearchRepositoryConstants.METADATA_DESCRIPTION,
								SearchRepositoryConstants.METADATA_KEYWORDS, SearchRepositoryConstants.METADATA_TITLE)
						.type(MultiMatchQueryBuilder.Type.PHRASE_PREFIX)));
		if (fullTextBuilder.hasClauses()) {
			queryBuilder.must(fullTextBuilder);
		}

		BoolQueryBuilder plBuilder = QueryBuilders.boolQuery();
		forEachTerm(searchInputDTO.getMetadata().getProgrammingLanguage(), term -> plBuilder
				.should(QueryBuilders.prefixQuery(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES, term)));
		if (plBuilder.hasClauses()) {
			queryBuilder.must(plBuilder);
		}

		BoolQueryBuilder keywordBuilder = QueryBuilders.boolQuery();
		forEachTerm(searchInputDTO.getMetadata().getKeyword(), term -> keywordBuilder
				.should(QueryBuilders.prefixQuery(SearchRepositoryConstants.METADATA_KEYWORDS, term)));
		if (keywordBuilder.hasClauses()) {
			queryBuilder.must(keywordBuilder);
		}

		BoolQueryBuilder formatBuilder = QueryBuilders.boolQuery();
		forEachTerm(searchInputDTO.getMetadata().getFormat(), term -> formatBuilder
				.should(QueryBuilders.prefixQuery(SearchRepositoryConstants.METADATA_FORMATS, term)));
		if (formatBuilder.hasClauses()) {
			queryBuilder.must(formatBuilder);
		}

		String parentId = searchInputDTO.getMetadata().getParentId();
		SearchResultDTO parentExercise = null;
		if (!StringUtils.isEmpty(parentId)) {
			try {
				parentExercise = getExerciseById(parentId, user);
			} catch (NotFoundException e) {
				String stripped = convertToPreviousId(parentId);
				if (stripped.equals(parentId)) {
					try {
						parentExercise = getExerciseById(stripped, user);
					} catch (NotFoundException e1) {
						LOGGER.warn("Cannot find exercise for: {}", parentId);
					}
				}
			}
			if (parentExercise == null || parentExercise.getFile().getChildren() == null) {
				// just a very lazy workaround
				return null;
			} else {
				String[] childrenIds = parentExercise.getFile().getChildren();
				BoolQueryBuilder collectionBuilder = QueryBuilders.boolQuery();
				for (String childId : childrenIds) {
					collectionBuilder.should(QueryBuilders.termQuery(SearchRepositoryConstants.EXERCISE_ID, childId));
					String previousId = convertToPreviousId(childId);
					if (!previousId.equals(childId)) {
						collectionBuilder
								.should(QueryBuilders.termQuery(SearchRepositoryConstants.EXERCISE_ID, previousId));
					}
				}
				queryBuilder.must(collectionBuilder);
			}
		}

		BoolQueryBuilder authorBuilder = QueryBuilders.boolQuery();
		if (searchInputDTO.getMetadata().getAuthor() != null) {
			authorBuilder.should(QueryBuilders
					.multiMatchQuery(searchInputDTO.getMetadata().getAuthor(),
							SearchRepositoryConstants.METADATA_CREATOR, SearchRepositoryConstants.METADATA_CONTRIBUTOR,
							SearchRepositoryConstants.METADATA_PUBLISHER)
					.type(MultiMatchQueryBuilder.Type.PHRASE_PREFIX));
		}
		if (authorBuilder.hasClauses()) {
			queryBuilder.must(authorBuilder);
		}

		if (searchInputDTO.getMetadata().getNaturalLanguage() != null
				&& !searchInputDTO.getMetadata().getNaturalLanguage().isEmpty()) {
			BoolQueryBuilder nlBuilder = QueryBuilders.boolQuery();
			searchInputDTO.getMetadata().getNaturalLanguage().forEach(language -> nlBuilder
					.should(QueryBuilders.matchQuery(SearchRepositoryConstants.METADATA_LANGUAGE, language)));
			if (nlBuilder.hasClauses()) {
				queryBuilder.must(nlBuilder);
			}
		}

		forEachTerm(searchInputDTO.getMetadata().getLicense(),
				term -> queryBuilder.must(QueryBuilders.prefixQuery(SearchRepositoryConstants.METADATA_LICENSE, term)));

		addAuthorizationQuery(user, queryBuilder);

		char[] boundaryChars = { '\n' };

		HighlightBuilder highlightBuilder = new HighlightBuilder().highlighterType("plain").preTags(preTags)
				.postTags(postTags).field(SearchRepositoryConstants.METADATA_DESCRIPTION).boundaryChars(boundaryChars)
				.boundaryScannerType(HighlightBuilder.BoundaryScannerType.SENTENCE).fragmentSize(FRAGMENT_SIZE);

		SearchSourceBuilder sourceBuilder = new SearchSourceBuilder();
		sourceBuilder.query(queryBuilder).highlighter(highlightBuilder).size(searchInputDTO.getPageSize())
				.from((searchInputDTO.getPage()) * searchInputDTO.getPageSize());

		SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_METADATA);
		searchRequest.source(sourceBuilder);

		SearchResponse searchResponse;
		try {
			searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);
		} catch (ElasticsearchException ese) {
			LOGGER.warn("Exception in elastic search", ese);
			throw ese;
		}

		return parseSearchResponse(searchResponse);
	}

	/**
	 * strips brackets from exercise Id. Just for temporary use. Will not be
	 * necessary, when we change to bracketed ids in general.
	 * 
	 * @param exerciseId
	 * @return
	 */
	public String convertToPreviousId(String exerciseId) {
		Pattern newParser = Pattern.compile("^\\[(\\d+)\\](.*)$");
		Matcher matcher = newParser.matcher(exerciseId.trim());
		if (matcher.matches()) {
			String projectId = matcher.group(1);
			String path = matcher.group(2);
			if (StringUtils.isEmpty(path)) {
				return projectId;
			} else {
				return projectId + ":" + path;
			}
		}
		return exerciseId;
	}

	/**
	 * this adds extra authorization requirements. It currently supports directly
	 * the group membership in the hierarchy and the new access management by shared
	 * groups and explicit members.
	 * 
	 * @param user
	 * @param queryBuilder
	 */
	private void addAuthorizationQuery(Optional<User> user, BoolQueryBuilder queryBuilder) {
		// Authorization restrictions
		final BoolQueryBuilder publicQuery = createPublicQuery();

		if (user.isEmpty()) {
			queryBuilder.must(publicQuery);
		} else {
			final Collection<GrantedAuthority> authorities = user.get().getAuthorities();
			final BoolQueryBuilder authQuery = QueryBuilders.boolQuery().boost(0.0f);
			// old membership management (can be eliminated some day)
			final Stream<QueryBuilder> prefixAuthQueries = authorities.stream()
					.filter(auth -> !Authority.isStandardRole(auth.getAuthority()))
					.filter(auth -> !auth.getAuthority().startsWith("SCOPE")).map(auth -> QueryBuilders
							.termQuery(SearchRepositoryConstants.PROJECT_NAMESPACE, auth.getAuthority()));
			prefixAuthQueries.forEach(pq -> authQuery.should(pq));

			// new membership management
			final Stream<QueryBuilder> prefixAuthQueries2 = authorities.stream()
					.filter(auth -> !Authority.isStandardRole(auth.getAuthority()))
					.filter(auth -> !auth.getAuthority().startsWith("SCOPE")).map(auth -> QueryBuilders
							.termQuery(SearchRepositoryConstants.PROJECT_GROUPS, auth.getAuthority()));
			prefixAuthQueries2.forEach(authQuery::should);
			final String email = user.get().getUsername().toLowerCase(Locale.getDefault());
			if (StringUtils.isNotEmpty(email)) {
				authQuery.should(QueryBuilders.termQuery(SearchRepositoryConstants.PROJECT_USERS, email));
			}
			if (authQuery.hasClauses()) {
				queryBuilder.must(QueryBuilders.boolQuery().should(authQuery).should(publicQuery));
			} else {
				queryBuilder.must(publicQuery);
			}
		}
	}

	private BoolQueryBuilder createPublicQuery() {
		final TermQueryBuilder simplePublicQuery = QueryBuilders
				.termQuery(SearchRepositoryConstants.PROJECT_VISIBILITY, "public").boost(0.0f);
		final ExistsQueryBuilder publicVisibilityQuery = QueryBuilders
				.existsQuery(SearchRepositoryConstants.METADATA_PUBLICVISIBILITY);

		return QueryBuilders.boolQuery().should(simplePublicQuery).should(publicVisibilityQuery);
	}

	private static SearchResultsDTO parseSearchResponse(final SearchResponse searchResponse)
			throws JsonProcessingException {
		float maxScore = searchResponse.getHits().getMaxScore();
		ObjectMapper objectMapper = new ObjectMapper();
		objectMapper.registerModule(new JavaTimeModule());
		objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

		final List<SearchResultDTO> searchResults = new ArrayList<>();
		for (SearchHit searchHit : searchResponse.getHits().getHits()) {
			float hitScore = searchHit.getScore();
			final SearchResultDTO parsedSearchHit = parseSearchHit(searchHit, objectMapper);
			int ranking5 = 1;
			if (maxScore > 0.0f)
				{ ranking5 = (int) ((hitScore * 5.0) / maxScore + 0.5f); }<>
			parsedSearchHit.setRanking5(ranking5);
			searchResults.add(parsedSearchHit);
		}

		long hitCount = searchResponse.getHits().getTotalHits().value;
		final SearchResultsDTO results = new SearchResultsDTO();
		results.setHitCount(hitCount);
		results.setSearchResult(searchResults);
		return results;
	}

	private static SearchResultDTO parseSearchHit(SearchHit searchHit, ObjectMapper objectMapper)
			throws JsonProcessingException {
		final SearchResultDTO mappedValue = objectMapper.readValue(searchHit.getSourceAsString(),
				SearchResultDTO.class);
		mappedValue.setExerciseId(searchHit.getId());
		return mappedValue;
	}

	/** Helper **/
	private static void forEachTerm(String searchTerm, Consumer<String> exec) {
		if (searchTerm == null) {
			return;
		}
		StringTokenizer st = new StringTokenizer(searchTerm);
		while (st.hasMoreTokens()) {
			exec.accept(st.nextToken());
		}
	}

	public SearchResultsDTO getExercisesById(Stream<String> exerciseIds, Optional<User> user, int page, int pageSize) {
		return getExercisesById(exerciseIds, user, page, pageSize, null);
	}

	/**
	 * returns all exercises, referred by their ids, changed after since (if not
	 * null)
	 * 
	 * @param exerciseIds a stream of exercise ids
	 * @param user        the user: needed for authorization
	 * @param page        the page
	 * @param pageSize    the page size
	 * @param since       an instant since when the exercise was last changed (or
	 *                    null, if not relevant)
	 * @return
	 */
	public SearchResultsDTO getExercisesById(Stream<String> exerciseIds, Optional<User> user, int page, int pageSize,
			Instant since) {
		BoolQueryBuilder idQueryBuilder = QueryBuilders.boolQuery();
		exerciseIds.forEach(exerciseId -> idQueryBuilder
				.should(QueryBuilders.matchQuery(SearchRepositoryConstants.EXERCISE_ID, exerciseId)));
		// Böse Falle: Wenn exerciseIds empty, dann wird alles zurückgegeben!
		if (!idQueryBuilder.hasClauses()) {
			final SearchResultsDTO results = new SearchResultsDTO();
			results.setHitCount(0L);
			results.setSearchResult(new ArrayList<>());
			return results;
		}
		queryBuilder.must(idQueryBuilder);
		if (since != null) {
			queryBuilder.must(QueryBuilders.rangeQuery(SearchRepositoryConstants.PROJECT_LASTACTIVITYAT).gte(since)
			/* .lte(Instant.now()) */
			);
		}

		BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery();

		addAuthorizationQuery(user, queryBuilder);

		SearchSourceBuilder sourceBuilder = new SearchSourceBuilder();
		sourceBuilder.query(queryBuilder).size(pageSize).from(page * pageSize);

		SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_METADATA);
		searchRequest.source(sourceBuilder);

		try {
			SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);

			return parseSearchResponse(searchResponse);
		} catch (JsonProcessingException e) {
			throw new NotFoundException(" Exercise with ids " + idQueryBuilder.toString() + " not found?", e);
		} catch (IOException e) {
			throw new NotFoundException("Query for exercise with id " + idQueryBuilder.toString() + " went wrong?", e);
		}
	}

	/**
	 * Searches for an exercise by id just a convenience method
	 * 
	 * @param exerciseId the parsed exercise id
	 * @param user       the user (if authenticated)
	 * @return search result
	 * @throws NotFoundException if not found
	 */
	@Cacheable(value = EXERCISE_BY_ID_CACHE, key = "((#user.present)?#user.get().username:'') + #exerciseId.toString()")
	public SearchResultDTO getExerciseById(ExerciseId exerciseId, Optional<User> user) {
		return getExerciseById(exerciseId.toString(), user);
	}

	/**
	 * Searches for an exercise by id
	 * 
	 * @param exerciseId exercise id as String
	 * @param user       the user (if authenticated)
	 * @return search result
	 * @throws NotFoundException if not found
	 */
	@Cacheable(value = EXERCISE_BY_ID_CACHE, key = "((#user.present)?#user.get().username:'') + #exerciseId.toString()")
	public SearchResultDTO getExerciseById(String exerciseId, Optional<User> user) {
		SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_METADATA);

		BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery();

		queryBuilder.must(QueryBuilders.matchQuery(SearchRepositoryConstants.EXERCISE_ID, exerciseId));

		SearchSourceBuilder sourceBuilder = new SearchSourceBuilder();
		sourceBuilder.query(queryBuilder).size(10);
		addAuthorizationQuery(user, queryBuilder);

		searchRequest.source(sourceBuilder);

		try {
			SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);

			ObjectMapper objectMapper = new ObjectMapper();
			objectMapper.registerModule(new JavaTimeModule());
			objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

			if (searchResponse.getHits().getTotalHits().value > 1) {
				LOGGER.warn("Found {} hits for exercise id {}, expected only one!",
						searchResponse.getHits().getTotalHits().value, exerciseId);
			}
			if (searchResponse.getHits().getHits().length > 0) {
				final SearchResultDTO parseSearchHit = parseSearchHit(searchResponse.getHits().getHits()[0],
						objectMapper);
				parseSearchHit.setRanking5(5);
				return parseSearchHit;
			}
		} catch (JsonProcessingException e) {
			throw new NotFoundException(" Exercise with id " + exerciseId + " not found?", e);
		} catch (IOException e) {
			throw new NotFoundException("Query for exercise with id " + exerciseId + " went wrong?", e);
		}
		// temporary for backward compatibility!
		String oldExerciseId = convertToPreviousId(exerciseId);
		if (oldExerciseId.equals(exerciseId)) {
			throw new NotFoundException(" Exercise with id " + exerciseId + " not found?");
		} else {
			return getExerciseById(oldExerciseId, user);
		}
	}
}
