package at.ac.uibk.gitsearch.security.oauth2;

import at.ac.uibk.gitsearch.domain.Authority;
import at.ac.uibk.gitsearch.service.dto.AdminUserDTO;
import java.time.Instant;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.function.Consumer;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.gitlab4j.api.Constants.TokenType;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.models.User;
import org.springframework.security.oauth2.client.authentication.OAuth2LoginAuthenticationToken;
import org.springframework.security.oauth2.client.registration.ClientRegistration;
import org.springframework.security.oauth2.core.OAuth2AccessToken;
import org.springframework.security.oauth2.core.oidc.user.OidcUser;
import org.springframework.stereotype.Component;

@Component
public class UserDetailsFetcher {

    private static final Logger logger = LogManager.getLogger(UserDetailsFetcher.class);

    /** fills in data from OIDC repository.
     *   returns true, if user data really changed **/
    public boolean updateUserDetails(AdminUserDTO u, OidcUser oidcUser, OAuth2LoginAuthenticationToken oAuth2LoginAuthenticationToken) {
        final OAuth2AccessToken accessToken = oAuth2LoginAuthenticationToken.getAccessToken();
        final ClientRegistration clientRegistration = oAuth2LoginAuthenticationToken.getClientRegistration();
        return updateUserDetails(u, oidcUser, accessToken, clientRegistration);
    }

    /**
     * fills in date from OIDC repository
     * @param u
     * @param oidcUser
     * @param accessToken
     * @param clientRegistration
     * @return
     */
    public boolean updateUserDetails(
        AdminUserDTO u,
        OidcUser oidcUser,
        final OAuth2AccessToken accessToken,
        final ClientRegistration clientRegistration
    ) {
        String idToken = accessToken.getTokenValue();
        String issuerURI = (String) clientRegistration.getProviderDetails().getConfigurationMetadata().get("issuer");
        try (GitLabApi gitLabApi = new GitLabApi(issuerURI, TokenType.OAUTH2_ACCESS, idToken)) {
            boolean modified = false;
            gitLabApi.enableRequestResponseLogging();
            // List<Project> memberProjects = gitLabApi.getProjectApi().getMemberProjects();
            final Set<String> authorities = new HashSet<>();
            // remove all old gitlab roles
            u.getAuthorities().stream().filter(Authority::isStandardRole).forEach(authorities::add);

            @SuppressWarnings("unchecked")
            final List<String> gitLabGroups = (List<String>) oidcUser.getAttribute("groups");
            // add gitlab groups
            authorities.addAll(reduceGroups(gitLabGroups)); // only add main groups

            User gitUser = gitLabApi.getUserApi().getCurrentUser();

            modified |= updateAttribute(gitUser.getAvatarUrl(), u.getImageUrl(), u::setImageUrl);
            modified |= updateAttribute(gitUser.getEmail(), u.getEmail(), u::setEmail);
            modified |= updateAttribute(gitUser.getName(), u.getLastName(), u::setLastName);
            if (!u.getAuthorities().equals(authorities)) {
                u.setAuthorities(authorities);
                modified = true;
            }
            // modified |= updateAttribute(gitUser.getUsername(),u.getLogin(), u::setLogin);

            if (modified) {
            	u.setLastModifiedDate(Instant.now());
            }
            // testMetaDataGenerator.generateTestData(gitLabApi, issuerURI, idToken);
            return modified;
        } catch (GitLabApiException e) {
            logger.warn("Cannot fetch details for user {}", oidcUser.getName(), e);
        }
        return false;
    }

    /**
     * unfortunatelly gitlab returns not only the main groups but also all groups in the project hierarchy. That makes the list
     * very large. We need only the main groups!
     * @param groups
     * @return
     */
    protected List<String> reduceGroups(List<String> groups) {
        List<String> unNormalizedGroups = new ArrayList<>(groups.size());
        unNormalizedGroups.addAll(groups);
        final Iterator<String> unNormalizedGroupsIterator = unNormalizedGroups.iterator();
        while (unNormalizedGroupsIterator.hasNext()) {
            String groupToCheck = unNormalizedGroupsIterator.next();
            for (String otherGroup : unNormalizedGroups) {
                if (
                    groupToCheck.startsWith(otherGroup) &&
                    groupToCheck.length() != otherGroup.length() && // same group, just continue
                    groupToCheck.charAt(otherGroup.length()) == '/'
                ) { // it is a real subgroup
                    unNormalizedGroupsIterator.remove();
                    break;
                }
            }
        }
        return unNormalizedGroups; // now normalized :-)
    }

    private static boolean updateAttribute(String newString, String oldString, Consumer<String> setter) {
        if (oldString == null) {
            if (newString == null) {
            	return false;
            }
        } else if (oldString.equals(newString)) {
            return false;
        }
        setter.accept(newString);
        return true;
    }
}
