package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.es.model.ArtemisExerciseInfo;
import at.ac.uibk.gitsearch.properties.ApplicationProperties;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.MapperFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import com.fasterxml.jackson.dataformat.yaml.YAMLGenerator;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Objects;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.WebTarget;
import net.minidev.json.JSONObject;
import net.minidev.json.JSONStyle;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.lib.Repository;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.models.Group;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;

@Service
public class ExerciseService {

    private final Logger log = LoggerFactory.getLogger(ExerciseService.class);

    private final ApplicationProperties applicationProperties;

    private final FileService fileService;

    private final GitlabService gitlabService;

    public ExerciseService(
        ApplicationProperties applicationProperties,
        FileService fileService,
        GitlabService gitlabService
    ) {
        this.applicationProperties = applicationProperties;
        this.fileService = fileService;
        this.gitlabService = gitlabService;
    }

    /**
     * Retrieve a zipped Artemis Exercise from a given URL
     * and store it in a temporary directory
     * @param exerciseUrl the url of the exposed exercise
     * @throws ArtemisImportError
     */
    @SuppressWarnings("PMD.AvoidCatchingGenericException")
    public void importExerciseFromArtemis(String exerciseUrl) throws ArtemisImportError {
        String exerciseToken = exerciseUrl.split("/")[exerciseUrl.split("/").length - 1];
        Client client = ClientBuilder.newClient();

        WebTarget target = client.target(exerciseUrl);
        try (
            InputStream inputStream = target
                .request()
                .header("Authorization", getApiKeyOfRegisteredConnectorByUrl(new URI(exerciseUrl).getAuthority()))
                .accept(javax.ws.rs.core.MediaType.APPLICATION_OCTET_STREAM)
                .get(InputStream.class)) {

            ZipInputStream zipInput = new ZipInputStream(inputStream);
            storeZipInTmp(zipInput, exerciseToken);
        } catch (URISyntaxException e) {
            log.error("Unable to add apiKey to the import request for [{}]. Try to request resource without apiKey.", exerciseUrl);
        } catch (Exception e) {
            throw new ArtemisImportError(
                "ZipFile couldn't be retrieved from [" + URLDecoder.decode(exerciseUrl, StandardCharsets.UTF_8) + "]" + e.getMessage()
            );
        }
    }

    /**
     * Import exercise to Gitlab.
     * Sequentially executes following processes:
     * - Apply user defined exercise metadata
     * - Create repository in Gitlab's group defined by the user
     * - Locally checking out the new created repository
     * - Copy, commit and push files to the new repository
     *
     * @param exerciseInfo of the exercise to import
     * @param exerciseToken of the stored exercise (in the temporary directory) to import
     */
    public void importExercise(ArtemisExerciseInfo exerciseInfo, String exerciseToken, Integer gitlabGroupId)
        throws GitLabApiException, GitAPIException, IOException, ArtemisImportError {
        applyExerciseInfoChanges(exerciseInfo, exerciseToken);
        renameFiles(exerciseToken);

        gitlabService.setJGitDefaultAuth();
        Group group = gitlabService.getGroupById(gitlabGroupId);
        String repoUrl = gitlabService.createRepository(group, exerciseInfo.getTitle());
        final String tmpDirPropertyKey = "java.io.tmpdir";
        try (Repository repo = gitlabService.checkoutRepo(repoUrl)) {
			gitlabService.commitAndPushToRepo(repo, new File(System.getProperty(tmpDirPropertyKey), exerciseToken));
            FileUtils.deleteDirectory(new File(System.getProperty(tmpDirPropertyKey), exerciseToken));
            FileUtils.deleteDirectory(repo.getDirectory().getParentFile());
        } catch (IOException e) {
            log.error("Could not remove directory: {} \n", e.getMessage());
        }
    }

    /**
     * Retrieves metadata from imported Artemis exercise
     * @param exerciseToken of the imported exercise
     * @return Content of exercise's 'metadata.yaml' and 'artemis.yaml'
     */
    public ArtemisExerciseInfo getArtemisExerciseInfo(String exerciseToken) throws IOException {
        ObjectMapper mapper = new ObjectMapper(new JsonFactory());
        mapper
            .enable(DeserializationFeature.READ_ENUMS_USING_TO_STRING)
            .enable(MapperFeature.ACCEPT_CASE_INSENSITIVE_ENUMS)
            .enable(DeserializationFeature.ACCEPT_SINGLE_VALUE_AS_ARRAY)
            .disable(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES);
        File exerciseDetailsFile = Objects.requireNonNull(
            new File(System.getProperty("java.io.tmpdir"), exerciseToken)
                .listFiles((file, name) -> name.startsWith("Exercise-Details-") && name.endsWith(".json"))
        )[0];
        mapArtemisExerciseInfoValues(exerciseDetailsFile);

        ArtemisExerciseInfo artemisExerciseInfo = mapper.readValue(exerciseDetailsFile, ArtemisExerciseInfo.class);

        writeArtemisMetadataFile(exerciseDetailsFile, artemisExerciseInfo);
        return artemisExerciseInfo;
    }

    /**
     * Utility method used to map Artemis Exercise metadata values to the ones used in
     * Sharing Metadata
     *
     * @param jsonFile to parse the values from
     * @throws IOException if an error occurs
     */
    private void mapArtemisExerciseInfoValues(File jsonFile) throws IOException {
        ObjectMapper mapper = new ObjectMapper();
        JSONObject jsonObject = mapper.readValue(jsonFile, JSONObject.class);
        if (jsonObject.containsKey("difficulty")) {
            switch (jsonObject.get("difficulty").toString()) {
                case "EASY":
                    jsonObject.put("difficulty", "SIMPLE");
                case "MEDIUM":
                    jsonObject.put("difficulty", "MEDIUM");
                case "HARD":
                    jsonObject.put("difficulty", "ADVANCED");
            }
        }

        if (!jsonObject.containsKey("metadataVersion")) {
            jsonObject.put("metadataVersion", "0.4");
        }

        if (!jsonObject.containsKey("license")) {
            jsonObject.put("license", "CC-SA-BY 4.0");
        }

        if (!jsonObject.containsKey("format")) {
            jsonObject.put("format", "artemis");
        }

        if (!jsonObject.containsKey("structure")) {
            jsonObject.put("structure", "atomic");
        }

        if (jsonObject.containsKey("instructors")) {
            jsonObject.put("creator", jsonObject.get("instructors"));
            jsonObject.remove("instructors");
        }

        if ("programming".equals(jsonObject.get("type").toString())) {
            jsonObject.put("type", "programming exercise");
        }

        try (FileWriter fileWriter = new FileWriter(jsonFile, StandardCharsets.UTF_8)) {
            fileWriter.write(jsonObject.toString(JSONStyle.NO_COMPRESS));
        }
    }

    /**
     * Utility method used to extract Artemis metadata into a 'artemis.yaml' file
     * @param jsonMetadataFile where the exercise metadata is stored
     * @param artemisExerciseInfo the entity of the stored metadata
     * @throws IOException if an error occurs
     */
    private void writeArtemisMetadataFile(File jsonMetadataFile, ArtemisExerciseInfo artemisExerciseInfo) throws IOException {
        ObjectMapper mapper = new ObjectMapper(
            new YAMLFactory().enable(YAMLGenerator.Feature.INDENT_ARRAYS).disable(YAMLGenerator.Feature.WRITE_DOC_START_MARKER)
        );

        File artemisMetadataFile = new File(jsonMetadataFile.getParent(), "artemis.yaml");
        if (!artemisMetadataFile.exists()) {
            boolean success = artemisMetadataFile.createNewFile();
            if (!success) {
                log.warn("Cannot create file");
            }
        }
        mapper.writeValue(artemisMetadataFile, artemisExerciseInfo.getArtemisInfoOnly());
    }

    /**
     * Used to temporary store content of Artemis' exercises Zip-Files into a tmp-subdirectory
     *
     * Recursion is used in case of nested Zip-Files
     *
     * @param zipInputStream containing the Zip-File to unzip
     * @param subDir name of the tmp-subdirectory
     */
    @SuppressWarnings("PMD.AssignmentInOperand")
    private void storeZipInTmp(ZipInputStream zipInputStream, String subDir) throws IOException {
        Path targetDir;

        if (subDir.startsWith(System.getProperty("java.io.tmpdir"))) {
            targetDir = Files.createDirectories(Paths.get(subDir));
        } else {
            targetDir = Files.createDirectories(Paths.get(System.getProperty("java.io.tmpdir"), subDir));
        }

        ZipEntry zipEntry;

        while ((zipEntry = zipInputStream.getNextEntry()) != null) {
            if (zipEntry.getName().contains(".git")) { // important as Artemis exercises come in submodules
                continue;
            }

            Path zipEntryPath = Paths.get(targetDir.toString(), zipEntry.getName());

            if (isZipFile(zipEntry)) {
                writeFile(zipEntryPath.toFile(), zipInputStream);
                String newSubDir;
                if (StringUtils.endsWithAny(zipEntry.getName(), new String[] { "-exercise.zip", "-solution.zip", "-tests.zip" })) {
                    newSubDir = Paths.get(targetDir.getParent().toString(), getRepoNameFromZipEntry(zipEntry)).toString();
                } else {
                    newSubDir =
                        Paths.get(targetDir.toString(), zipEntry.getName().substring(0, zipEntry.getName().lastIndexOf("."))).toString();
                }

                storeZipInTmp(new ZipInputStream(new FileInputStream(zipEntryPath.toFile())), newSubDir);
                if (Files.deleteIfExists(zipEntryPath)) {
                    log.warn("Unable to delete temporary Zip file [{}] after extraction. Skipping..", zipEntryPath);
                }
            } else if (zipEntry.isDirectory()) {
                Files.createDirectory(zipEntryPath);
            } else {
                writeFile(zipEntryPath.toFile(), zipInputStream);
            }
        }
        cleanUpTmpDirectory(targetDir.toFile());
        zipInputStream.close();
    }

    /**
     * Utility method used to check if a ZipEntry represents a Zip file
     * @param zipEntry to check
     * @return true if zipEntry is a Zip file, false otherwise
     */
    private boolean isZipFile(ZipEntry zipEntry) {
        return zipEntry.getName().endsWith(".zip");
    }

    /**
     * Cleans up a temporary directory by removing it if empty
     * or scheduling its deletion with a delay of 60 minutes, otherwise.
     *
     * @param tmpDir to clean up
     */
    private void cleanUpTmpDirectory(File tmpDir) {
        try {
            if (tmpDir.exists() && tmpDir.isDirectory() && fileService.isEmpty(tmpDir.toPath())) {
                Files.delete(tmpDir.toPath());
            }
            fileService.scheduleForDeletion(tmpDir.toPath(), 60);
        } catch (IOException e) {
            log.error("An error occurred while cleaning up a temporary directory: ", e);
        }
    }

    /**
     * Utility method used to extract the repository name from Zip files
     * representing git submodules. (Artemis exercise exports come in
     * submodules for 'exercise', 'solution' and 'test')
     * @param zipEntry to get the repository name from
     * @return the repository name
     */
    private String getRepoNameFromZipEntry(ZipEntry zipEntry) {
        return zipEntry.getName().substring(zipEntry.getName().lastIndexOf("-") + 1, zipEntry.getName().lastIndexOf("."));
    }

    /**
     * Method used to write files from a zipInputStream
     * @param entryFile File to write
     * @param zipInputStream Byte stream to read from
     * @throws IOException
     */
    @SuppressWarnings("PMD.AssignmentInOperand")
    private void writeFile(File entryFile, ZipInputStream zipInputStream) throws IOException {
        byte[] buffer = new byte[1024];

        if (!entryFile.getParentFile().exists()) {
            Files.createDirectories(entryFile.getParentFile().toPath());
        }

        try (FileOutputStream fos = new FileOutputStream(entryFile)) {
            int len;
            while ((len = zipInputStream.read(buffer)) > 0) {
                fos.write(buffer, 0, len);
            }
        }
    }

    /**
     * Utility method used to apply metadata changes to the exercise's 'metadata.yaml' file
     * @param exerciseInfo containing the metadata values
     * @param exerciseToken of the exercise to update
     * @throws IOException
     */
    public void applyExerciseInfoChanges(ArtemisExerciseInfo exerciseInfo, String exerciseToken) throws IOException {
        ObjectMapper mapper = new ObjectMapper(
            new YAMLFactory().enable(YAMLGenerator.Feature.INDENT_ARRAYS).disable(YAMLGenerator.Feature.WRITE_DOC_START_MARKER)
        );
        mapper.writeValue(
            Paths.get(System.getProperty("java.io.tmpdir"), exerciseToken, "metadata.yaml").toFile(),
            exerciseInfo.getExerciseInfoOnly()
        );
    }

    /**
     * Used to validate apiKeys by checking if they match with
     * one of the registeredConnectors' keys stored in application-(dev|prod).yml
     * @param apiKey the apikey to validate
     * @return true if equal to the local one, false otherwise
     */
    public boolean validate(String apiKey) {
        for (ApplicationProperties.RegisteredConnector registeredConnector : applicationProperties.getRegisteredConnectors()) {
            if (registeredConnector.getAccessToken().equals(apiKey)) {
                log.debug("ApiKey validated correctly against [" + registeredConnector.getUrl() + "]");
                return true;
            }
        }
        return false;
    }

    /**
     * Method that retrieves the ApiKey of a Registered Connector configured in 'application-dev/prod.yml'
     * given the URL of the connector
     * @param urlAuthority URL Authority/Domain of the connector
     * @return the ApiKey, or Null if no connector with the given url is found
     */
    private String getApiKeyOfRegisteredConnectorByUrl(String urlAuthority) {
        for (ApplicationProperties.RegisteredConnector registeredConnector : applicationProperties.getRegisteredConnectors()) {
            if (registeredConnector.getUrl().contains(urlAuthority)) {
                return registeredConnector.getAccessToken();
            }
        }
        return null;
    }

    /**
     * Renames specific files of an imported Artemis exercise,
     * to meet naming conventions
     *
     * Renamed files:
     * - Problem-Statement-xxx.md -> exercise.md
     * - Exercise-Details-xxx.json -> Exercise-Details.json
     *
     * @param exerciseToken to retrieve the location of the files to rename
     */
    private void renameFiles(String exerciseToken) {
        try {
            File problemStatementFile = Objects.requireNonNull(
                new File(System.getProperty("java.io.tmpdir"), exerciseToken)
                    .listFiles((file, name) -> name.startsWith("Problem-Statement-") && name.endsWith(".md"))
            )[0];
            if (!problemStatementFile.renameTo(Paths.get(System.getProperty("java.io.tmpdir"), exerciseToken, "exercise.md").toFile())) {
                log.error("Unable to rename file [{}]. Skipping..", problemStatementFile.getName());
            }
            File exerciseDetailsFile = Objects.requireNonNull(
                new File(System.getProperty("java.io.tmpdir"), exerciseToken)
                    .listFiles((file, name) -> name.startsWith("Exercise-Details-") && name.endsWith(".json"))
            )[0];
            Files.move(exerciseDetailsFile.toPath(), exerciseDetailsFile.toPath().resolveSibling("Exercise-Details.json"));
        } catch (ArrayIndexOutOfBoundsException e) {
            log.error("File to rename could not be found in [{}].", System.getProperty("java.io.tmpdir") + File.separator + exerciseToken);
        } catch (IOException ioe) {
            log.error("Error while renaming file: {}", ioe.getMessage());
        }
    }
}
