package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Collectors;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;
import org.apache.commons.io.FileUtils;
import org.apache.http.auth.AuthenticationException;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.util.ExerciseId;
import org.eclipse.jgit.api.Git;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.lib.Repository;
import org.eclipse.jgit.transport.CredentialsProvider;
import org.eclipse.jgit.transport.UsernamePasswordCredentialsProvider;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.GroupApi;
import org.gitlab4j.api.ProjectApi;
import org.gitlab4j.api.UserApi;
import org.gitlab4j.api.models.Group;
import org.gitlab4j.api.models.Project;
import org.gitlab4j.api.models.RepositoryFile;
import org.gitlab4j.api.models.Visibility;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.core.userdetails.User;
import org.springframework.stereotype.Service;
import org.springframework.util.StreamUtils;
import org.springframework.util.StringUtils;

/**
 * Service for accessing git lab backbone
 * <p>
 */
@Service
public class GitlabService {

    @Autowired
    @SuppressWarnings("PMD.ImmutableField")
    private GitLabRepository gitLabRepository;

    @Autowired
    private MetaDataRepository metaDataRepository;

    @Autowired
    protected TokenProvider tokenProvider;

    @Autowired
    private FileService fileService;

    private static final Logger log = LoggerFactory.getLogger(ShoppingBasketService.class);

    public Boolean repositoryExists(String projectID) {
        try {
            final GitLabApi gitLabApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo());
            final ProjectApi gitLabProjectApi = gitLabApi.getProjectApi();
            return gitLabProjectApi.getProject(projectID) != null;
        } catch (GitLabApiException e) {
            log.error(e.getMessage(), e);
            return false;
        }
    }

    /**
     * returns an authentication of the current user against gitlab, together with groups and access token
     */
    public Authentication getGitLabAuthentication(String name, String oAuth2Token) {
        final GitLabApi gitLabApi = gitLabRepository.getGitLabApi(oAuth2Token);
        List<Group> groups = Collections.emptyList();
        if (!org.apache.commons.lang3.StringUtils.isEmpty(oAuth2Token)) { try {
            groups = gitLabApi.getGroupApi().getGroups();
        } catch (GitLabApiException e) {
            log.warn("Cannot read groups for {}", name);
            groups = Collections.emptyList();
        }}

        Collection<? extends GrantedAuthority> authorities = groups
            .stream()
            .map(g -> g.getFullPath())
            .map(SimpleGrantedAuthority::new)
            .collect(Collectors.toList());

        User principal = new User(name, "", authorities);

        final UsernamePasswordAuthenticationToken authentication = new UsernamePasswordAuthenticationToken(principal, "", authorities);
        Map<String, String> detailsMap = new HashMap<>();
        detailsMap.put(TokenProvider.GITLAB_ACCESS_TOKEN, oAuth2Token);
        detailsMap.put(TokenProvider.GITLAB_ACCESS_ISSUER, "who cares?");
        authentication.setDetails(detailsMap);
        return authentication;
    }

    public InputStream getRepositoryZip(String exerciseID) throws GitLabApiException, IOException {
        final GitLabApi gitLabApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo());
        return rePackageGitLabProjectZip(
            new ZipInputStream(gitLabApi.getRepositoryApi().getRepositoryArchive(exerciseID, "HEAD", "zip")),
            "from project " + exerciseID
        );
    }

    /**
     * chops of the main folder, and brings every file one level up. Also deleting
     * all plain files in main folder.
     *
     * @param zipIn the zip to be repackaged
     * @param originalLocation the original location. For debug purposes only.
     * @return
     * @throws IOException
     */
    protected InputStream rePackageGitLabProjectZip(ZipInputStream zipIn, String originalLocation) throws IOException {
        final PipedInputStream pis = new PipedInputStream(1024 * 256);
        final PipedOutputStream pos = new PipedOutputStream(pis);

        Runnable rePackage = () -> {
            try (ZipOutputStream zipOut = new ZipOutputStream(pos)) {
                ZipEntry zipInEntry = zipIn.getNextEntry();
                String prefix = "";
                if (zipInEntry.isDirectory()) { // main directory is prefix to all entries
                    prefix = zipInEntry.getName();
                }

                while (zipInEntry != null) {
                    if (zipInEntry.getName().startsWith(prefix)) {
                        String newName = zipInEntry.getName().substring(prefix.length());
                        if (StringUtils.hasLength(newName)) {
                            ZipEntry zipOutEntry = new ZipEntry(newName);
                            if (zipInEntry.isDirectory()) {
                                log.debug("ignoring directory {}", zipInEntry.getName());
                                zipOut.putNextEntry(zipOutEntry);
                            } else {
                                zipOut.putNextEntry(zipOutEntry);
                                StreamUtils.copy(zipIn, zipOut);
                                zipOut.closeEntry();
                            }
                        }
                    }
                    zipInEntry = zipIn.getNextEntry();
                }
                zipIn.close();
            } catch (IOException e) {
                log.error("Cannot rezip file from {}", originalLocation);
            }
        };

        new Thread(rePackage).start();

        return pis;
    }

    /**
     * returns the requested file from the project in gitlab. If the user is not enabled for access, but the project is public visible, the file is returned with admin token.
     * @param exerciseId
     * @param filePath
     * @return
     * @throws GitLabApiException
     * @throws IOException
     */
    public InputStream getRepositoryFile(ExerciseId exerciseId, String filePath) throws GitLabApiException, IOException {
        final SearchResultDTO exercise = metaDataRepository.getExerciseById(exerciseId.toString(), tokenProvider.getCurrentPrincipal());
        if (exercise == null) {
        	throw new IOException("Exercise with id " + exerciseId + " not found!");
        }

        final GitLabApi gitLabApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo());
        RepositoryFile file;
        try {
            file =
                gitLabApi
                    .getRepositoryFileApi()
                    .getFile(exercise.getProject().getProjectId(), exerciseId.extendPath(filePath), exercise.getFile().getCommit_id());
        } catch (GitLabApiException e) {
            log.warn("Cannot read file {}", e.getMessage());
            if (
                exercise.getMetadata().getPublicVisibility() != null &&
                (!isDisabledPrefix(filePath, exercise.getMetadata().getPublicVisibility().getExcept()))
            ) {
                file =
                    gitLabRepository
                        .getAdminGitLabApi()
                        .getRepositoryFileApi()
                        .getFile(exercise.getProject().getProjectId(), exerciseId.extendPath(filePath), exercise.getFile().getCommit_id());
            } else {
                throw new IOException("Access to " + filePath + " not allowed");
            }
        }
        return new ByteArrayInputStream(file.getDecodedContentAsBytes());
    }

    private boolean isDisabledPrefix(String filePath, String[] exceptions) {
        for (String exception : exceptions) {
            if (
                filePath.startsWith(exception) && (filePath.equals(exception) || filePath.charAt(exception.length() + 1) == '/')
            ) {
            	return true;
            }
        }
        return false;
    }

    /**
     * Creates a new repository in Gitlab and returns the url to it
     * @param group where the repository should be created
     * @param repoName of the repository
     * @return Url (as String) to the repository
     * @throws GitLabApiException
     */
    public String createRepository(Group group, String repoName) throws GitLabApiException, ArtemisImportError {
        final ProjectApi projectApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo()).getProjectApi();

        try {
            // If repository already exist, return it
            return projectApi.getProject(group.getFullPath() + "/" + repoName.replace(" ", "-")).getWebUrl();
        } catch (GitLabApiException ignored) {
            log.debug("not relevant here" + ignored.getMessage());
        }

        Project project = new Project().withNamespaceId(group.getId()).withName(repoName).withVisibility(Visibility.PUBLIC);

        try {
            project = projectApi.createProject(project);
        } catch (GitLabApiException e) {
            log.error("An error occurred while creating the new Gitlab project : {}", e.getMessage());
            throw new ArtemisImportError("exercise.import.error.creationFailed");
        }

        return project.getWebUrl();
    }

    public void commitAndPushToRepo(Repository repository, File srcDir) throws IOException, ArtemisImportError {
        Git git = new Git(repository);

        try {
            final UserApi userApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo()).getUserApi();
            org.gitlab4j.api.models.User currentUser = userApi.getCurrentUser();

            FileUtils.copyDirectory(srcDir, git.getRepository().getDirectory().getParentFile());
            git.add().addFilepattern(".").call();
            git.commit().setAll(true).setMessage("Importing files").setCommitter(currentUser.getUsername(), currentUser.getEmail()).call();
            git.push().setCredentialsProvider(CredentialsProvider.getDefault()).call();
        } catch (GitLabApiException | GitAPIException e) {
            if (git.getRepository().getDirectory().getParentFile().exists()) {
                try {
                    FileUtils.deleteDirectory(git.getRepository().getDirectory().getParentFile());
                } catch (IOException ioe) {
                    log.error("Could not remove directory: {} \n", ioe.getMessage());
                }
            }
            log.error("An error occurred during the commit and push of files: {}", e.getMessage());
            git.getRepository().close();
            git.close();
            throw new ArtemisImportError("exercise.import.error.pushFailed");
        }

        git.getRepository().close();
        git.close();
    }

    /**
     * Checkout repository in the local 'tmp' directory
     * @param repoUrl of the repository to checkout
     * @return a JGit Repository entity
     */
    public Repository checkoutRepo(String repoUrl) throws ArtemisImportError {
        String repoDirName = repoUrl.split("/")[repoUrl.split("/").length - 1];
        File dst = new File(System.getProperty("java.io.tmpdir"), repoDirName);
        if (dst.exists()) {
            log.debug("Temporary directory '{}' already exists, removing it.", dst.getPath());
            try {
                FileUtils.deleteDirectory(dst);
            } catch (IOException e) {
                log.error("Unable to remove directory. Create another directory : {} \n", e.getMessage());
                dst = new File(System.getProperty("java.io.tmpdir"), repoDirName + UUID.randomUUID());
            }
        }
        try {
            Git git = Git.cloneRepository().setURI(repoUrl).setDirectory(dst).call();
            Repository repo = git.getRepository();
            repo.close();
            git.close();

            fileService.scheduleForDeletion(repo.getDirectory().toPath(), 60);
            return repo;
        } catch (GitAPIException e) {
            log.error("Unable to clone Git repository to temporary directory : {}", e.getMessage());
            throw new ArtemisImportError("exercise.import.error.checkoutFailed");
        }
    }

    /**
     * Retrieve Group by its Id
     * @param groupId of the group to retrieve
     * @return Gitlab Group
     * @throws GitLabApiException if an error occurs
     */
    public Group getGroupById(Integer groupId) throws GitLabApiException {
        GroupApi groupApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo()).getGroupApi();

        return groupApi.getGroup(groupId);
    }

    /**
     * Used to retrieve Gitlab groups
     * (As user: their groups, as admin: all groups)
     * @return List of all groups
     * @throws GitLabApiException if an error occurs
     */
    public List<Group> getGroups() throws GitLabApiException, AuthenticationException {
        if (tokenProvider.getGitLabAccessInfo().isEmpty()) {
            throw new AuthenticationException();
        }
        GroupApi groupApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo()).getGroupApi();

        User principal = tokenProvider.getCurrentPrincipal().orElseThrow(AuthenticationException::new);
        Collection<String> authorities = principal
            .getAuthorities()
            .stream()
            .map(GrantedAuthority::getAuthority)
            .collect(Collectors.toList());

        return groupApi.getGroups().stream().filter(group -> authorities.contains(group.getFullPath())).collect(Collectors.toList());
    }

    /**
     * Used to retrieve all Gitlab groups where the current user
     * has at least a MAINTAINER membership
     *
     * @return List of all groups with at least MAINTAINER membership
     * @throws AuthenticationException
     * @throws GitLabApiException
     */
    public List<Group> getAllGroupsWithAtLeastMaintainerRole() throws AuthenticationException, GitLabApiException {
        UserApi userApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo()).getUserApi();
        GroupApi groupApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo()).getGroupApi();
        Integer userId = userApi.getCurrentUser().getId();

        List<Group> groups = getGroups();
        groups =
            groups
                .stream()
                .filter(group -> {
                    try {
                        return groupApi.getMember(group.getId(), userId, true).getAccessLevel().value >= 40;
                    } catch (GitLabApiException e) {
                        log.warn("Could not retrieve membership of user [{}] for group [{}]. Skipping..", userId, group.getId());
                    }
                    return false;
                })
                .collect(Collectors.toList());

        return groups;
    }

    /**
     * Used to retrieve Gitlab subgroups of a given group
     * @param groupId of the group to search in
     * @return List of subgroups of given group
     * @throws GitLabApiException if an error occurs
     */
    public List<Group> getSubGroups(int groupId) throws GitLabApiException {
        GroupApi groupApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo()).getGroupApi();

        return groupApi.getSubGroups(groupId);
    }

    /**
     * Utility function used to set the default JGIT credentials
     * to use the provided OAuth2 token
     */
    public void setJGitDefaultAuth() {
        final Optional<String> gitLabAccessToken = tokenProvider.getGitLabAccessToken();
        if (gitLabAccessToken.isPresent()) {
            CredentialsProvider.setDefault(new UsernamePasswordCredentialsProvider("oauth2", gitLabAccessToken.get()));
        } else {
            log.error("Cannot find gitlab token!");
        }
    }
}
