package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.domain.User;
import at.ac.uibk.gitsearch.domain.UserWatchList;
import at.ac.uibk.gitsearch.domain.enumeration.CheckFrequency;
import at.ac.uibk.gitsearch.repository.UserRepository;
import java.nio.charset.StandardCharsets;
import java.time.Instant;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;
import javax.mail.MessagingException;
import javax.mail.internet.MimeMessage;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.MessageSource;
import org.springframework.mail.MailException;
import org.springframework.mail.javamail.JavaMailSender;
import org.springframework.mail.javamail.MimeMessageHelper;
import org.springframework.scheduling.annotation.Async;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;
import org.thymeleaf.context.Context;
import org.thymeleaf.spring5.SpringTemplateEngine;
import tech.jhipster.config.JHipsterProperties;

/**
 * Service for sending emails.
 * <p>
 * We use the {@link Async} annotation to send emails asynchronously.
 */
@Service
public class MailService {

    private final Logger log = LoggerFactory.getLogger(MailService.class);

    private static final String USER = "user";

    private static final String BASE_URL = "baseUrl";

    private final JHipsterProperties jHipsterProperties;

    private final JavaMailSender javaMailSender;

    private final MessageSource messageSource;

    private final SpringTemplateEngine templateEngine;

    private final UserRepository userRepository;

    private final UserWatchListService userWatchListService;

    public MailService(
        JHipsterProperties jHipsterProperties,
        JavaMailSender javaMailSender,
        MessageSource messageSource,
        SpringTemplateEngine templateEngine,
        UserRepository userRepository,
        UserWatchListService userWatchListService
    ) {
        this.jHipsterProperties = jHipsterProperties;
        this.javaMailSender = javaMailSender;
        this.messageSource = messageSource;
        this.templateEngine = templateEngine;
        this.userRepository = userRepository;
        this.userWatchListService = userWatchListService;
    }

    /**
     * send email
     * @param to
     * @param subject
     * @param content
     * @param isMultipart
     * @param isHtml
     * @return true if mail was successfully sent.
     */
    @Async
    public boolean sendEmail(String to, String subject, String content, boolean isMultipart, boolean isHtml) {
        log.debug(
            "Send email[multipart '{}' and html '{}'] to '{}' with subject '{}' and content={}",
            isMultipart,
            isHtml,
            to,
            subject,
            content
        );

        // Prepare message using a Spring helper
        MimeMessage mimeMessage = javaMailSender.createMimeMessage();
        try {
            MimeMessageHelper message = new MimeMessageHelper(mimeMessage, isMultipart, StandardCharsets.UTF_8.name());
            message.setTo(to);
            message.setFrom(jHipsterProperties.getMail().getFrom());
            message.setSubject(subject);
            message.setText(content, isHtml);
            javaMailSender.send(mimeMessage);
            log.debug("Sent email to User '{}'", to);
            return true;
        } catch (MailException | MessagingException e) {
            log.warn("Email could not be sent to user '{}'", to, e);
            return false;
        }
    }

    @Async
    public void sendEmailFromTemplate(User user, String templateName, String titleKey) {
        if (user.getEmail() == null) {
            log.debug("Email doesn't exist for user '{}'", user.getLogin());
            return;
        }
        Locale locale = Locale.forLanguageTag(user.getLangKey());
        Context context = new Context(locale);
        context.setVariable(USER, user);
        context.setVariable(BASE_URL, jHipsterProperties.getMail().getBaseUrl());
        String content = templateEngine.process(templateName, context);
        String subject = messageSource.getMessage(titleKey, null, locale);
        sendEmail(user.getEmail(), subject, content, false, true);
    }

    /**
     * Send info mails to every user, iff he/she was scheduled for sending.
     * This is scheduled to get fired everyday, at 05:00 (am).
     */
    @Scheduled(cron = "0 0 5 * * ?")
    public void sendInfoMails() {
        for (User user : userRepository.findAll()) {
            sendInfoMail(user, false/* only if content */);
        }
    }

    /**
     * send an info mail to user, if relevant.
     * For test purposes mainly.
     * @param user the user
     * @param force force mail sending, even if mail is empty.
     * @return true if mail was sent.
     */
    public boolean sendInfoMail(User user, boolean force) {
    	User sendingUser = user;
        boolean hasContent = force;
        Locale locale = Locale.forLanguageTag(sendingUser.getLangKey());
        Context context = new Context(locale);
        context.setVariable(USER, sendingUser);
        context.setVariable(BASE_URL, jHipsterProperties.getMail().getBaseUrl());
        Instant lastMailSending = sendingUser.getLastMailSent();
        if (lastMailSending == null) {
            lastMailSending = sendingUser.getLastLogin();
            if (lastMailSending == null) {
                lastMailSending = Instant.ofEpochMilli(0); // should only happen during testing
            }
            sendingUser.setLastMailSent(lastMailSending);
            sendingUser = userRepository.save(sendingUser);
        }
        Map<SearchResultDTO, UserWatchList> findChangesSince = null;

        findChangesSince = userWatchListService.findChangesSince(lastMailSending, sendingUser);

        boolean dailyRelevant = findChangesSince
            .entrySet()
            .stream()
            .anyMatch(entry -> entry.getValue().getCheckFrequency() == CheckFrequency.DAILY);
        boolean weeklyRelevant = findChangesSince
            .entrySet()
            .stream()
            .anyMatch(entry -> entry.getValue().getCheckFrequency() == CheckFrequency.WEEKLY);
        boolean monthlyRelevant = findChangesSince
            .entrySet()
            .stream()
            .anyMatch(entry -> entry.getValue().getCheckFrequency() == CheckFrequency.MONTHLY);

        boolean shouldSend =
            (dailyRelevant && lastMailSending.isBefore(CheckFrequency.DAILY.getChangedBefore())) ||
            (weeklyRelevant && lastMailSending.isBefore(CheckFrequency.WEEKLY.getChangedBefore())) ||
            (monthlyRelevant && lastMailSending.isBefore(CheckFrequency.MONTHLY.getChangedBefore()));

        hasContent = hasContent || shouldSend;

        if (hasContent) {
            final List<Entry<SearchResultDTO, UserWatchList>> changedExercises = findChangesSince
                .entrySet()
                .stream()
                .collect(Collectors.toList());
            context.setVariable("watchListUpdates", changedExercises);
            String content = templateEngine.process("mail/info/infoEmail", context);
            String subject = messageSource.getMessage("email.info.title", null, locale);
            boolean success = sendEmail(sendingUser.getEmail(), subject, content, false, true);
            if (success) {
                sendingUser.setLastMailSent(Instant.now());
                userRepository.save(sendingUser);
                return true;
            }
        }
        return false;
    }

    @Async
    public void sendActivationEmail(User user) {
        log.debug("Sending activation email to '{}'", user.getEmail());
        sendEmailFromTemplate(user, "mail/activationEmail", "email.activation.title");
    }

    @Async
    public void sendCreationEmail(User user) {
        log.debug("Sending creation email to '{}'", user.getEmail());
        sendEmailFromTemplate(user, "mail/creationEmail", "email.activation.title");
    }

    @Async
    public void sendPasswordResetMail(User user) {
        log.debug("Sending password reset email to '{}'", user.getEmail());
        sendEmailFromTemplate(user, "mail/passwordResetEmail", "email.reset.title");
    }
}
