package at.ac.uibk.gitsearch.web.rest;

import at.ac.uibk.gitsearch.repository.StatisticsRepository;
import at.ac.uibk.gitsearch.service.SearchService;
import at.ac.uibk.gitsearch.service.StatisticsService;
import at.ac.uibk.gitsearch.service.dto.StatisticsDTO;
import at.ac.uibk.gitsearch.web.rest.errors.BadRequestAlertException;
import at.ac.uibk.gitsearch.web.rest.utils.RestUtils;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.List;
import java.util.Optional;
import javax.servlet.http.HttpServletRequest;
import javax.validation.Valid;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpHeaders;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.servlet.HandlerMapping;
import org.springframework.web.servlet.support.ServletUriComponentsBuilder;
import tech.jhipster.web.util.HeaderUtil;
import tech.jhipster.web.util.PaginationUtil;
import tech.jhipster.web.util.ResponseUtil;

/**
 * REST controller for managing {@link at.ac.uibk.gitsearch.domain.Statistics}.
 */
@RestController
@RequestMapping("/api")
public class StatisticsResource {

    private static final Logger log = LoggerFactory.getLogger(StatisticsResource.class);

    private static final String ENTITY_NAME = "statistics";

    //  @Value("${jhipster.clientApp.name}") // warning: inconsistent rename of gitSearchApp to gitSearchV2App :-(
    private static final String applicationName = "gitsearchApp";

    private final StatisticsService statisticsService;

    @Autowired
    private final SearchService searchService;

    private final StatisticsRepository statisticsRepository;

    public StatisticsResource(StatisticsService statisticsService, SearchService searchService, StatisticsRepository statisticsRepository) {
        this.statisticsService = statisticsService;
        this.searchService = searchService;
        this.statisticsRepository = statisticsRepository;
    }

    /**
     * {@code POST  /statistics} : Create a new statistics.
     *
     * @param statisticsDTO the statisticsDTO to create.
     * @return the {@link ResponseEntity} with status {@code 201 (Created)} and with
     *         body the new statisticsDTO, or with status {@code 400 (Bad Request)}
     *         if the statistics has already an ID.
     * @throws URISyntaxException if the Location URI syntax is incorrect.
     */
    @PostMapping("/statistics")
    @SuppressWarnings("PMD")
    @PreAuthorize("hasAnyRole('ADMIN')")
    public ResponseEntity<StatisticsDTO> createStatistics(@Valid @RequestBody StatisticsDTO statisticsDTO) throws URISyntaxException {
        log.debug("REST request to save Statistics : {}", statisticsDTO);
        if (statisticsDTO.getId() != null) {
            throw new BadRequestAlertException("A new statistics cannot already have an ID", ENTITY_NAME, "idexists");
        }
        StatisticsDTO result = statisticsService.save(statisticsDTO);
        return ResponseEntity
            .created(new URI("/api/statistics/" + result.getId()))
            .headers(HeaderUtil.createEntityCreationAlert(applicationName, true, ENTITY_NAME, result.getId().toString()))
            .body(result);
    }

    /**
     * {@code PUT  /statistics} : Updates an existing statistics.
     *
     * @param statisticsDTO the statisticsDTO to update.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and with body
     *         the updated statisticsDTO, or with status {@code 400 (Bad Request)}
     *         if the statisticsDTO is not valid, or with status
     *         {@code 500 (Internal Server Error)} if the statisticsDTO couldn't be
     *         updated.
     * @throws URISyntaxException if the Location URI syntax is incorrect.
     */
    @PutMapping("/statistics")
    @PreAuthorize("hasAnyRole('ADMIN')")
    public ResponseEntity<StatisticsDTO> updateStatistics(@Valid @RequestBody StatisticsDTO statisticsDTO) throws URISyntaxException {
        log.debug("REST request to update Statistics : {}", statisticsDTO);
        if (statisticsDTO.getId() == null) {
            throw new BadRequestAlertException("Invalid id", ENTITY_NAME, "idnull");
        }
        StatisticsDTO result = statisticsService.save(statisticsDTO);
        return ResponseEntity
            .ok()
            .headers(HeaderUtil.createEntityUpdateAlert(applicationName, true, ENTITY_NAME, statisticsDTO.getId().toString()))
            .body(result);
    }

    /**
     * {@code GET  /statistics} : get all the statistics.
     *
     * @param pageable the pagination information.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and the list
     *         of statistics in body.
     */
    @GetMapping("/statistics")
    @PreAuthorize("hasAnyRole('ADMIN')")
    public ResponseEntity<List<StatisticsDTO>> getAllStatistics(Pageable pageable) {
        log.debug("REST request to get a page of Statistics");
        Page<StatisticsDTO> page = statisticsService.findAll(pageable);
        HttpHeaders headers = PaginationUtil.generatePaginationHttpHeaders(ServletUriComponentsBuilder.fromCurrentRequest(), page);
        return ResponseEntity.ok().headers(headers).body(page.getContent());
    }

    /**
     * {@code GET  /statistics/:id} : get the "id" statistics.
     *
     * @param statisticId the id of the statisticsDTO to retrieve.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and with body
     *         the statisticsDTO, or with status {@code 404 (Not Found)}.
     */
    @GetMapping("/statistics/{statisticId}")
    @PreAuthorize("hasAnyRole('ADMIN')")
    public ResponseEntity<StatisticsDTO> getStatistics(@PathVariable Long statisticId) {
        log.debug("REST request to get Statistics : {}", statisticId);
        Optional<StatisticsDTO> statisticsDTO = statisticsService.findOne(statisticId);
        return ResponseUtil.wrapOrNotFound(statisticsDTO);
    }

    /*
    getStatisticsByExerciseId is used to match the exerciseId of a Gitlab project with the database object for the statistics
    If a rest request for an exerciseId comes from the client and there is no object with a fitting exerciseId in the DB the server will create a new db entry with number of views
    as 1 and downloads as 0 
    */

    @GetMapping("/statistics/exercise/**")
    public ResponseEntity<StatisticsDTO> getStatisticsByExerciseId(HttpServletRequest request) {
        String pattern = (String) request.getAttribute(HandlerMapping.BEST_MATCHING_PATTERN_ATTRIBUTE);

        String exerciseId = RestUtils.decodeURL(request, pattern);

        log.debug("REST request to get exercise Statistics for ExerciseID : {}", exerciseId);

        Optional<StatisticsDTO> statisticsDTO = statisticsService.findOneByExerciseID(exerciseId);

        if (statisticsDTO.isPresent()) {
            StatisticsDTO newStats = statisticsDTO.get();
            newStats.setViews(newStats.getViews() + 1);
            statisticsService.save(newStats);
            return ResponseUtil.wrapOrNotFound(Optional.of(newStats));
        }

        // TODO there is a race condition, if serveral parallel requests do not find an entry.
        if (searchService.findExerciseById(exerciseId).isPresent()) {
            // the exercise exists in the meta data search index
            StatisticsDTO newStats = new StatisticsDTO();
            newStats.setDownloads(0);
            newStats.setViews(1);
            newStats.setExerciseID(exerciseId);
            statisticsService.save(newStats);
            log.debug("Created new statistics entry for exerciseID: {}", exerciseId);
            statisticsDTO = Optional.of(newStats);
            return ResponseUtil.wrapOrNotFound(statisticsDTO);
        }
        return ResponseUtil.wrapOrNotFound(statisticsDTO);
    }

    /**
     * {@code DELETE  /statistics/:id} : delete the "id" statistics.
     *
     * @param id the id of the statisticsDTO to delete.
     * @return the {@link ResponseEntity} with status {@code 204 (NO_CONTENT)}.
     */
    @DeleteMapping("/statistics/{id}")
    @PreAuthorize("hasAnyRole('ADMIN')")
    public ResponseEntity<Void> deleteStatistics(@PathVariable Long id) {
        log.debug("REST request to delete Statistics : {}", id);
        statisticsService.delete(id);
        return ResponseEntity
            .noContent()
            .headers(HeaderUtil.createEntityDeletionAlert(applicationName, true, ENTITY_NAME, id.toString()))
            .build();
    }

    //    /**
    //     * {@code SEARCH  /_search/statistics?query=:query} : search for the statistics
    //     * corresponding to the query.
    //     *
    //     * @param query    the query of the statistics search.
    //     * @param pageable the pagination information.
    //     * @return the result of the search.
    //     */
    //    @GetMapping("/_search/statistics")
    //    @PreAuthorize("hasAnyRole('ADMIN')")
    //    public ResponseEntity<List<StatisticsDTO>> searchStatistics(@RequestParam String query, Pageable pageable) {
    //        log.debug("REST request to search for a page of Statistics for query {}", query);
    //        Page<StatisticsDTO> page = statisticsService.search(query, pageable);
    //        HttpHeaders headers = PaginationUtil
    //                .generatePaginationHttpHeaders(ServletUriComponentsBuilder.fromCurrentRequest(), page);
    //        return ResponseEntity.ok().headers(headers).body(page.getContent());
    //    }

    @GetMapping("/statistics/numberOfWatchlistEntries/**")
    public ResponseEntity<Integer> getNumberOfWatchListEntriesForExerciseID(HttpServletRequest request) {
        String pattern = (String) request.getAttribute(HandlerMapping.BEST_MATCHING_PATTERN_ATTRIBUTE);

        String exerciseId = RestUtils.decodeURL(request, pattern);

        log.debug("REST request to get number of watchlists for project {}", exerciseId);
        Integer n = statisticsRepository.findNumberOfWatchListEntriesByExerciseID(exerciseId);
        return ResponseEntity.ok().body(n);
    }
}
